/*-
 * Copyright 2008 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef __CX88_AUDIO_IOCTL_H__
#define __CX88_AUDIO_IOCTL_H__

#include "cx88_common_ioctl.h"

#define CX88_AUDIO_SAMPLE_SIZE 4

typedef struct
{  
   #define CX88_AUDIO_FMT_FM_RADIO     0
   #define CX88_AUDIO_FMT_FM_RADIO_50  1
   #define CX88_AUDIO_FMT_FM_RADIO_75  2
   #define CX88_AUDIO_FMT_BTSC_FM      3
   #define CX88_AUDIO_FMT_BTSC         4
   #define CX88_AUDIO_FMT_EIAJ         5
   #define CX88_AUDIO_FMT_AM_L         6
   #define CX88_AUDIO_FMT_A2_BG        7
   #define CX88_AUDIO_FMT_A2_DK        8
   #define CX88_AUDIO_FMT_A1_I         9
   #define CX88_AUDIO_FMT_NICAM_BGDK   10
   #define CX88_AUDIO_FMT_NICAM_I      11
   #define CX88_AUDIO_FMT_NICAM_L      12
   uint8_t in_audio_format;
   
   #define CX88_AUDIO_STEREO_OFF       0
   #define CX88_AUDIO_STEREO_ON        1
   #define CX88_AUDIO_STEREO_AUTO      2
   uint8_t in_stereo_format;
   
   uint8_t in_second_lang;
   
   #define CX88_AUDIO_INPUT_I2S             0
   #define CX88_AUDIO_INPUT_BCAST           1
   #define CX23885_AUDIO_INPUT_BCAST2       2
   #define CX23885_AUDIO_INPUT_BCAST3       3
   #define CX23885_AUDIO_INPUT_BCAST4       4
   #define CX23885_AUDIO_INPUT_BCAST5       5
   #define CX23885_AUDIO_INPUT_AC97         6
   #define CX23885_AUDIO_INPUT_AF           (1 << 3)
   #define CX23885_AUDIO_INPUT_ADC1         (1 << 4)
   #define CX23885_AUDIO_INPUT_ADC2         (2 << 4)
   #define CX23885_AUDIO_INPUT_DEMOD_STEREO (1 << 6)
   #define CX23885_AUDIO_INPUT_DEMOD_MONO   (2 << 6)
   uint8_t in_input;
      
   #define CX88_AUDIO_I2S_IN_MASTER     (1 << 0)
   #define CX88_AUDIO_I2S_IN_SONY       (1 << 1)
   #define CX88_AUDIO_I2S_OUT_SONY      (1 << 2)
   #define CX88_AUDIO_I2S_TO_DAC        (1 << 3)
   #define CX23885_AUDIO_I2S_OUT_MASTER (1 << 3)
   #define CX88_AUDIO_I2S_MODE_OSRDAC   (1 << 4)
   #define CX88_AUDIO_I2S_MODE_OSRADC   (2 << 4)
   #define CX88_AUDIO_I2S_MODE_PASS     (3 << 4)
   #define CX88_AUDIO_I2S_MODE_FULL     (4 << 4)
   uint8_t in_i2s_config;
   
} cx88_audio_format;

typedef struct
{
   void *in_buffer; /* User-mode buffer allocation */
   uint32_t inout_num_samples; /* Number of samples to allocate */
   uint32_t out_samples_per_buffer; /* Number of packets in each buffer (last buffer may be smaller) */
   uint8_t out_num_buffers; /* Number of buffers */
   uint8_t in_perfect_fit; /* 0: driver can modify inout_num_samples to be slightly lower to optimize DMA efficiency */
} cx88_audio_config;

typedef struct
{
   uint32_t in_timeout_ms;  /* Maximum time to wait (0 -> no timeout) */   
   
   uint8_t out_buffer_available; /* 1-based index of most recent available buffer. 0->no buffers available */
   
   /* Capture status */
   #define CX88_AUDIO_CAP_NOERR     0
   #define CX88_AUDIO_CAP_FATAL     (1 << 0)
   #define CX88_AUDIO_CAP_TIMEOUT   (1 << 1)
   #define CX88_AUDIO_OVERFLOW      (1 << 2)
   #define CX88_AUDIO_BERTHRESH     (1 << 3)
   #define CX88_AUDIO_SYNC_ERROR    ((1 << 4) | CX88_AUDIO_CAP_FATAL)
   #define CX88_AUDIO_BUS_ERROR     ((1 << 5) | CX88_AUDIO_CAP_FATAL)
   #define CX88_AUDIO_DMA_PROGRAM   ((1 << 6) | CX88_AUDIO_CAP_FATAL)
   #define CX88_AUDIO_WAIT_ABORTED  ((1 << 7) | CX88_AUDIO_CAP_FATAL)
   uint8_t out_status;
} cx88_audio_wait;

typedef struct
{
   uint8_t in_left;
   uint8_t in_right;
} cx88_audio_volume;

typedef struct
{
   uint8_t out_vol_left;
   uint8_t out_vol_right;
   uint8_t out_mute;
   
   #define CX88_AUDIO_NO_CARRIER 0
   #define CX88_AUDIO_CARRIER_1  (1 << 0)
   #define CX88_AUDIO_CARRIER_2  (1 << 1)
   uint8_t out_carrier;
   
   #define CX88_AUDIO_SIGNAL_STEREO    0
   #define CX88_AUDIO_SIGNAL_DUAL_MONO 1
   #define CX88_AUDIO_SIGNAL_MONO      2
   #define CX88_AUDIO_SIGNAL_SAP       3
   uint8_t out_signal;
   
} cx88_audio_status;

#define CX88_AUDIO_FORMAT      _IOWR('A', 0, cx88_audio_format)
#define CX88_AUDIO_CONFIG      _IOWR('A', 1, cx88_audio_config)
#define CX88_AUDIO_CAPTURE     _IO('A', 2)
#define CX88_AUDIO_STOP        _IO('A', 3)
#define CX88_AUDIO_WAIT        _IOWR('A', 4, cx88_audio_wait)
#define CX88_AUDIO_VOLUME      _IOW('A', 5, cx88_audio_volume)
#define CX88_AUDIO_MUTE        _IO('A', 6)
#define CX88_AUDIO_GET_STATUS  _IOR('A', 7, cx88_audio_status)

#endif
