#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include "cx88_video_ioctl.h"
#include "tuner_devnode_device.h"

int main(int argc, char **argv)
{
   uint32_t id;
   if (argc < 2)
   {
      printf("Usage: %s <devnode>\n", argv[0]);
      return 1;  
   }
   int fd = open(argv[1], O_RDWR);
   int error = ioctl(fd, CX88_COMMON_CARD_ID, &id);
   if (error)
   {
      printf("Error during card ID ioctl\n");
      return 1;
   }
   printf("Card ID = 0x%x\n", id);
   cx88_common_i2c_config config;
   config.in_bus_number = 0;
   sprintf(config.devices[0].in_driver, "iicdev");
   config.devices[0].in_speed = CX88_I2C_SPEED_100KHZ;
   config.devices[0].in_addr = 0x50;
   config.in_num_i2c_devices = 1;
   error = ioctl(fd, CX88_COMMON_I2C_CONFIG, &config);
   if (error)
   {
      printf("Error during I2C config ioctl\n");
      return 1;  
   }
   char eeprom_devnode[16];
   sprintf(eeprom_devnode, "/dev/iicdev%d", config.devices[0].out_instance);
   tuner_config conf;
   tuner_devnode_device eeprom_dev(conf, eeprom_devnode, error);
   if (error)
   {
      printf("Error creating eeprom devnode\n");
      return 1;
   }
   uint8_t buffer[] = {0x00, 0x00};
   error = eeprom_dev.transact(buffer, 1, &(buffer[1]), 1);
   if (error)
   {
      printf("Error reading function enables\n");  
   }
   if (!(buffer[1] & 0x2))
   {
      printf("Audio function not detected: enabling\n");
      buffer[1] |= 0x2;
      size_t written = 0;
      error = eeprom_dev.write(buffer, sizeof(buffer), written);
      if (error || (written == 0))
      {
         printf("Error writing eeprom!\n");
         return 1;
      }
      sleep(10);
   }
   else
   {
      printf("Audio function already enabled: Function mask = 0x%x\n", buffer[1]);  
   }
   close(fd);
   return 0;
}

