#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include "cx88_video_ioctl.h"
#include "tuner_devnode_device.h"
#include "lgh064f.h"
#include "dtt7612.h"

extern char *optarg;

void capture(int fd, void *bufspace, cx88_video_format &fmt, cx88_video_config &config, cx88_video_capture &capture)
{
   printf("Beginning capture\n");
   int error = ioctl(fd, CX88_VIDEO_CAPTURE, &capture);
   if (error)
   {
      printf("Error during capture ioctl\n");
      return;
   }
   char filename[32];
   for (int i = 0; i < config.in_num_frames; ++i)
   {
      cx88_video_wait wait;
      wait.in_timeout_ms = 0;
      if ((i == 0) || (wait.channels[CX88_VIDEO_CHANNEL_Y].out_even_frame_available <= i))
      {
         wait.in_channel_mask = CX88_VIDEO_FIELD_EVEN(CX88_VIDEO_CHANNEL_Y);
         error = ioctl(fd, CX88_VIDEO_WAIT, &wait);
         if (error)
         {
            printf("Error during wait ioctl\n");
            ioctl(fd, CX88_VIDEO_STOP);
            return;
         }
         if ((wait.out_status & CX88_VIDEO_BUS_ERROR) == CX88_VIDEO_BUS_ERROR)
         {
            printf("Fatal error: PCI bus error\n");
         }
         if ((wait.out_status & CX88_VIDEO_SYNC_ERROR) == CX88_VIDEO_SYNC_ERROR)
         {
            printf("Fatal error[%d]: DMA sync error\n", i);  
         }
         if ((wait.out_status & CX88_VIDEO_DMA_PROGRAM) == CX88_VIDEO_DMA_PROGRAM)
         {
            printf("Fatal error[%d]: DMA program error\n", i);
         }
         if ((wait.out_status & CX88_VIDEO_WAIT_ABORTED) == CX88_VIDEO_WAIT_ABORTED)
         {
            printf("Fatal error: Wait aborted\n");
         }
         if ((wait.out_status & CX88_VIDEO_CAP_TIMEOUT) == CX88_VIDEO_CAP_TIMEOUT)
         {
            printf("Nonfatal error: Capture timed out\n");
         }
         if (wait.out_status & CX88_VIDEO_CAP_FATAL)
         {
            ioctl(fd, CX88_VIDEO_STOP);
            return;
         }
      }
      if ((wait.channels[CX88_VIDEO_CHANNEL_Y].out_even_frame_available <= i) ||
          (wait.channels[CX88_VIDEO_CHANNEL_Y].out_odd_frame_available <= i))
      {
         printf("Warning: No frame available %d %d\n",
            wait.channels[CX88_VIDEO_CHANNEL_Y].out_even_frame_available,
            wait.channels[CX88_VIDEO_CHANNEL_Y].out_odd_frame_available);
      }
      printf("VBI frames: %d, %d\n",
         wait.channels[CX88_VIDEO_CHANNEL_VBI].out_odd_frame_available,
         wait.channels[CX88_VIDEO_CHANNEL_VBI].out_even_frame_available);
      printf("U frames: %d, %d\n",
         wait.channels[CX88_VIDEO_CHANNEL_U].out_odd_frame_available,
         wait.channels[CX88_VIDEO_CHANNEL_U].out_even_frame_available);
      printf("V frames: %d, %d\n",
         wait.channels[CX88_VIDEO_CHANNEL_V].out_odd_frame_available,
         wait.channels[CX88_VIDEO_CHANNEL_V].out_even_frame_available);
      sprintf(filename, "out%d.ppm", i);
      FILE *outfile = fopen(filename, "w");
      if (outfile == NULL)
      {
         printf("Error opening video output file\n");
         ioctl(fd, CX88_VIDEO_STOP);
         return;  
      }
      uint32_t linesize = fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].line_size + config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_x_header + config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_x_trailer;
      uint32_t numlines = fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].num_lines + config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_y_header + config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_y_trailer;
      uint8_t *ybuf, *ubuf, *vbuf;

      ybuf = reinterpret_cast <uint8_t*> 
         (config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_buffers[i]);
      ubuf = reinterpret_cast <uint8_t*>  
         (config.bufconfig[CX88_VIDEO_CHANNEL_U].in_buffers[i]);
      vbuf = reinterpret_cast <uint8_t*>  
         (config.bufconfig[CX88_VIDEO_CHANNEL_V].in_buffers[i]);

      size_t ymult = 1, umult = 1, vmult = 1;
      if (fmt.in_pixel_format == CX88_VIDEO_YUV422)
      {
         ubuf = ybuf + 1;
         vbuf = ybuf + 3;
         ymult = 2;
         umult = 4;
         vmult = 4;
         linesize /= 2;
      }
      fprintf(outfile, "P6\n%d %d\n255\n", linesize, numlines);
      uint32_t bufsize = linesize * numlines;

      for (size_t j = 0; j < bufsize; ++j)
      {
         uint8_t pixel[3];
         int8_t y = ybuf[j * ymult];
         int8_t u = ubuf[(j / 2) * umult] - 128;
         int8_t v = vbuf[(j / 2) * vmult] - 128;
         double r = 1.370705 * v + ybuf[j * ymult];
         double g = (double)y - (0.337633 * u) - (0.698001 * v);
         double b = 1.732446 * u + y;
         if (r < 0) r = 0;
         if (g < 0) g = 0;
         if (b < 0) b = 0;
         if (r > 255) r = 255;
         if (g > 255) g = 255;
         if (b > 255) b = 255;
         pixel[0] = (uint8_t) r;
         pixel[1] = (uint8_t) g;
         pixel[2] = (uint8_t) b;
         fwrite(pixel, 3, 1, outfile);
      }
      /*for (size_t j = 0; j < bufsize; j += 3)
      {
         uint8_t pixel[3];
         pixel[0] = ybuf[j + 2];
         pixel[1] = ybuf[j + 1];
         pixel[2] = ybuf[j];
         fwrite(pixel, 3, 1, outfile);
      }*/
      fclose(outfile);
      if ((wait.channels[CX88_VIDEO_CHANNEL_VBI].out_even_frame_available <= i) ||
          (wait.channels[CX88_VIDEO_CHANNEL_VBI].out_odd_frame_available <= i))
      {
         printf("Warning: no VBI frame available\n");
         continue;
      }
      bufsize = fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].line_size * fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].num_lines;
      char *vbibuf = reinterpret_cast <char*> 
         (config.bufconfig[CX88_VIDEO_CHANNEL_VBI].in_buffers[i]);
      sprintf(filename, "out%d.vbi", i);
      FILE *vbifile = fopen(filename, "w");
      if (vbifile == NULL)
      {
         printf("Warning: unable to open VBI output file\n");
         continue;
      }
      fwrite(vbibuf, bufsize, 1, vbifile);
      fclose(vbifile);
   }
   ioctl(fd, CX88_VIDEO_STOP);
}

void tune(int fd, void *bufspace, cx88_video_format &fmt, cx88_video_config &config, cx88_video_capture &cap)
{
   if (cap.in_input_type == CX88_VIDEO_INPUT_SVIDEO)
   {
      capture(fd, bufspace, fmt, config, cap);
      return;
   }
   uint32_t id;
   int error = ioctl(fd, CX88_COMMON_CARD_ID, &id);
   if (error)
   {
      printf("Error during card ID ioctl\n");
      return;
   }
   printf("Card ID = 0x%x\n", id);
   cx88_common_i2c_config i2config;
   i2config.in_bus_number = 0;
   sprintf(i2config.devices[0].in_driver, "iicdev");
   i2config.devices[0].in_speed = CX88_I2C_SPEED_100KHZ;
   i2config.devices[0].in_addr = 0x61;
   i2config.in_num_i2c_devices = 1;
   error = ioctl(fd, CX88_COMMON_I2C_CONFIG, &i2config);
   if (error)
   {
      printf("Error during I2C config ioctl\n");
      return;  
   }
   char tuner_devnode[16];
   sprintf(tuner_devnode, "/dev/iicdev%d", i2config.devices[0].out_instance);
   printf("Tuner: %s\n", tuner_devnode);
   avb_channel channel;
   channel.video_format = AVB_VIDEO_FMT_NTSC_M;
   channel.audio_format = AVB_AUDIO_FMT_AUTO;
   channel.frequency_hz = 177000000;
   channel.bandwidth_hz = 6000000;
   tuner_config conf;
   error = conf.load_file("tuner_config");
   if (error)
   {
      printf("Error loading tuner config file\n");
      return;
   }
   tuner_devnode_device tuner_dev(conf, tuner_devnode, error);
   if (error)
   {
      printf("Error creating tuner devnode\n");
      return;
   }
   pll_driver *tuner_driver = NULL;
   switch (id)
   {
      case 0x70633000:
         tuner_driver = new dtt7612(conf, tuner_dev);
         break;
      case 0x70635500:
         tuner_driver = new lgh064f(conf, tuner_dev);
         break;
      default:
         printf("Unrecognized card type: %d\n", id);
         break;
   }
   if (tuner_driver == NULL) return;
   error = tuner_driver->set_channel(channel);
   if (error)
   {
      printf("Error setting channel on tuner driver\n");
      return;  
   }
   printf("Ready to start tuner\n");
   error = tuner_driver->start(10000);
   if (error)
   {
      printf("Error starting tuner\n");
      return;
   }
   sleep(1);
   capture(fd, bufspace, fmt, config, cap);
   delete tuner_driver;
}

int main(int argc, char **argv)
{
   cx88_video_format fmt;
   fmt.in_video_format = CX88_VIDEO_FMT_NTSCM;
   fmt.in_pixel_format = CX88_VIDEO_YCC422;
   fmt.in_rows = 480;
   fmt.in_cols = 640;
   fmt.in_x_offset = 0;
   fmt.in_x_size = 640;
   fmt.in_y_offset = 0;
   fmt.in_y_size = 480;
   fmt.in_enable_vbi = 1;
   fmt.in_field_options = CX88_VIDEO_FIELD_INTERLACE;
   fmt.in_even_field.swap = 0;
   fmt.in_even_field.enable = 1;
   fmt.in_odd_field.swap = 0;
   fmt.in_odd_field.enable = 1;
   cx88_video_config config;
   uint8_t num_frames = 0;
   int fd = open(argv[1], O_RDWR);
   uint32_t pixfmts = 0;
   int error = ioctl(fd, CX88_VIDEO_PIXFORMATS, &pixfmts);
   if (error)
   {
      printf("Error during pixformats ioctl: %s\n", strerror(errno));
      return 1;
   }
   if (!(pixfmts & CX88_VIDEO_YCC422)) fmt.in_pixel_format = CX88_VIDEO_YUV422;
   error = ioctl(fd, CX88_VIDEO_FORMAT, &fmt);
   if (error)
   {  
      printf("Error during format ioctl: %s\n", strerror(errno));
      return 1;
   }
   size_t xheader = 120;
   size_t xtrailer = 240;
   size_t yheader = 33;
   size_t ytrailer = 56;
   size_t y_size = (xheader + xtrailer + fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].line_size) * 
      (yheader + ytrailer + fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].num_lines);
   size_t u_size = (((xheader + xtrailer) / 2) + fmt.out_geoms[CX88_VIDEO_CHANNEL_U].line_size)
      * (yheader + ytrailer + fmt.out_geoms[CX88_VIDEO_CHANNEL_U].num_lines);
   size_t v_size = (((xheader + xtrailer) / 2) + fmt.out_geoms[CX88_VIDEO_CHANNEL_V].line_size) 
      * (yheader + ytrailer + fmt.out_geoms[CX88_VIDEO_CHANNEL_V].num_lines);
   size_t vbi_size = fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].line_size * fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].num_lines;
   void *buf = NULL;
   cx88_video_capture capture;
   capture.in_input_source = CX88_VIDEO_INPUT_MUX1;
   capture.in_input_type = CX88_VIDEO_INPUT_COMPOSITE;
   capture.in_input_options = 0;
   capture.in_lock_timeout_ms = 1000;
   capture.in_lock_delay_ms = 100;
   capture.in_no_frame_intrs = 0;
   int optch = -1;
   while ((optch = getopt(argc - 1, &(argv[1]), "u:s:")) != -1)
   {
      switch (optch)
      {
         case 'u':
            num_frames = strtoul(optarg, NULL, 10);
            buf = malloc(num_frames * (y_size + u_size + v_size + vbi_size));
            if (buf == NULL)
            {
               printf("Error allocating video buffer\n");
               goto cleanup;
            }
            break;
         case 's':
            capture.in_input_source = strtoul(optarg, NULL, 10);
            capture.in_input_type = CX88_VIDEO_INPUT_SVIDEO;
            break;
      }
   }
   if (num_frames <= 2)
   {
      printf("Please specify a number of frames to capture that is greater than 2, using -u or -k\n");
      goto cleanup;
   }
   for (uint8_t frames = num_frames; frames >= num_frames - 2; frames -= 2)
   {
      sleep(10);
      config.in_num_frames = frames;
      config.in_one_shot = 1;
      config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_y_header = yheader;
      config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_y_trailer = ytrailer;
      config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_x_header = xheader;
      config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_x_trailer = xtrailer;
      config.bufconfig[CX88_VIDEO_CHANNEL_U].in_y_header = yheader;
      config.bufconfig[CX88_VIDEO_CHANNEL_U].in_y_trailer = ytrailer;
      config.bufconfig[CX88_VIDEO_CHANNEL_U].in_x_header = xheader / 2;
      config.bufconfig[CX88_VIDEO_CHANNEL_U].in_x_trailer = xtrailer / 2;
      config.bufconfig[CX88_VIDEO_CHANNEL_V].in_y_header =  yheader;
      config.bufconfig[CX88_VIDEO_CHANNEL_V].in_y_trailer = ytrailer;
      config.bufconfig[CX88_VIDEO_CHANNEL_V].in_x_header = xheader / 2;
      config.bufconfig[CX88_VIDEO_CHANNEL_V].in_x_trailer = xtrailer / 2;
      config.bufconfig[CX88_VIDEO_CHANNEL_VBI].in_x_header = 0;
      config.bufconfig[CX88_VIDEO_CHANNEL_VBI].in_x_trailer = 0;
      config.bufconfig[CX88_VIDEO_CHANNEL_VBI].in_y_header = 0;
      config.bufconfig[CX88_VIDEO_CHANNEL_VBI].in_y_trailer = 0;
      for (uint8_t i = 0; i < frames; ++i)
      {
         config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_buffers[i] = (char*)buf + (i * y_size);
         config.bufconfig[CX88_VIDEO_CHANNEL_U].in_buffers[i] = (char*)buf + (num_frames * y_size) + (i * u_size);
         config.bufconfig[CX88_VIDEO_CHANNEL_V].in_buffers[i] = (char*)buf + (num_frames * (y_size + u_size)) + (i * v_size);
         config.bufconfig[CX88_VIDEO_CHANNEL_VBI].in_buffers[i] = (char*)buf + (num_frames * (y_size + u_size + v_size)) + (i * vbi_size);
      }

      printf("Y Geometry: %d x %d\n", fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].line_size, fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].num_lines);
      printf("U Geometry: %d x %d\n", fmt.out_geoms[CX88_VIDEO_CHANNEL_U].line_size, fmt.out_geoms[CX88_VIDEO_CHANNEL_U].num_lines);
      printf("V Geometry: %d x %d\n", fmt.out_geoms[CX88_VIDEO_CHANNEL_V].line_size, fmt.out_geoms[CX88_VIDEO_CHANNEL_V].num_lines);
      printf("VBI Geometry: %d x %d\n", fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].line_size, fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].num_lines);
      error = ioctl(fd, CX88_VIDEO_CONFIG, &config);
      if (error)
      {
         printf("Error during config ioctl: %s\n", strerror(errno));
         return 1;
      }
      tune(fd, buf, fmt, config, capture);
   }
cleanup:
   close(fd);
   free(buf);
   return 0;
}

